/**********************************************************************************/
/*                                                                                */ 
/* dcasvm_learn-0.03: Dual Coorinate Ascent SVM (ver. 0.03)                       */ 
/*                                                                                */ 
/* Author: Yutaka Sasaki, Toyota Technological Institute                          */ 
/*         (Yutaka.Sasaki<AT>toyota-ti.ac.jp)                                     */ 
/*                                                                                */ 
/* Licensed under the GNU General Public License, Version 2 (the "License")       */
/* You may not use this file except in compliance with the License.               */
/* You may obtain a copy of the License at                                        */
/*                                                                                */
/*      http://http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt               */
/*                                                                                */
/* Unless required by applicable law or agreed to in writing, software            */
/* distributed under the License is distributed on an "AS IS" BASIS,              */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.       */
/* See the License for the specific language governing permissions and            */
/* limitations under the License.                                                 */
/**********************************************************************************/

/* Disclaimer: This is not a good C++ code. This is C implementation with C++     */
/* data type extension. Minimum Object-Oriented progamming is involved.           */

/* Header files */
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <map>
#include <set>
#include <vector>

#define MAX_G 64
#define MAX_X 460000
#define MAX_CATEGORY 52
#define MAX_CINDEX 30
#define MAX_FEATURE 3000
#define MAX_FINDEX 2100000
#define FNAME_LEN 128
#define ROOT 2143406
/* #define ROOT 2244783 */

char version[] = " Dual Coordinate Ascent L1-SVM (ver. 0.03) Y. Sasaki, TTI, Japan\n";
int max_findex = 1000;
double*  W;
double B;
char* model_file;
char* test_file;
char* result_file="result.dcasvm";

std::vector< std::vector< std::pair< int, double > > >  X;
std::map< int, int > Y;

// /* prototypes */
int read_model_file ( char fname[] );
int dcal1_read_test_file ( char fname[] );
void usage( char* command );
double dot_product( int docid );
void read_option (int argc, char *argv[]);
void dcal1_test( int N, char* fname );

int main(int argc, char *argv[]) {

  int n_data;
  int i;
  char* command;

  printf( "------------------------------------------------------------------\n" );
  printf( version );
  printf( "------------------------------------------------------------------\n" );

  command = argv[0];
  read_option( argc, argv );
  if ( (model_file=='\0')||( test_file=='\0') ) { 
        usage( command );
        exit(0); 
  }

  printf( "-----------------------------------------\n" );
  printf( "| Test data file name = %s\n", test_file );
  printf( "| Model file name = %s\n", model_file );
  printf( "| Result file name = %s\n", result_file );
  printf( "-----------------------------------------\n" );
  n_data = dcal1_read_test_file ( test_file );
  fprintf( stderr, "%d samples read.\n", n_data );

  /* test the SVM model */
  W = new double[max_findex+1];
  for( i=0; i<= max_findex ; i++ ) { W[i]=0; }

  read_model_file ( model_file );  

  dcal1_test( n_data, result_file );

}

void dcal1_test( int N, char* fname ) {
  FILE *rfd;
  double g;

  printf( "Writing result file \"%s\" ... \n", fname);
  rfd = fopen( fname, "w" );
  if ( rfd == NULL ) { return; }

  for ( int i=0; i<N ; i++ ) {
     g = dot_product(i)+B;
     fprintf( rfd, "%g\t%d\n", g, Y[i] );
  }

  fclose( rfd );
  printf( "Done.\n" );
}

int read_model_file ( char fname[] ) {

  //  FILE *mfd;

  int fsize;
  int i,j;
  int count_f;
  char c;
  char satt[100];
  char sval[100];
  int iatt;
  double fval;
  std::string strbuf;
  char* buf;
  std::ifstream ifs;

  printf( "Reading model file \"%s\" ... \n", fname);

  ifs.open( fname );
  if ( ifs.fail() ) { printf( "Not found!\n" ); exit(1); }

  // Read b
  std::getline(ifs,strbuf);
  fprintf( stderr, "buf size =  %d\n", (int)strbuf.size() );
  buf = new char[strbuf.size()+1];
  strcpy( buf, strbuf.c_str() );
  //  printf( "[%s]\n",buf);

  fsize = (int)strbuf.size()+1;
  fprintf( stderr, "file size =  %d\n", fsize );

  i=0;
  j=0;
  c = buf[i++];
  while ( c != ' ' ) {
       sval[j++] = c;
        c = buf[i++];
  }
  sval[j]='\0';
  B = atof( sval );

  count_f = 0;
  for( ; i<=fsize ; ) {
     
        j=0;
        c = buf[i++];
        if ( c=='#' ) { break; }
       	while ( c != ':' ) {
          satt[j++] = c;
          c = buf[i++];
        }
	satt[j]='\0';
        iatt = atoi( satt );

        j=0;
        c = buf[i++];
	while ( (c != ' ') && (c != '\n') && ( i<=fsize ) ){
          sval[j++] = c;
          c = buf[i++];
        }
	sval[j]='\0';
        fval = atof( sval );
	//	printf( "[%d] %d:%f\n", i, iatt, fval );
	if ( iatt <= max_findex ) W[iatt] = fval;
        count_f++;
  } /* end of for */
  //  fclose( mfd );
  delete[] buf;
  ifs.close();
  printf( "Done.\n" );
  fprintf( stderr, "\n");

  return 1;
}

int dcal1_read_test_file ( char fname[] ) {

  int iatt, icat=1;
  int line=0;
  register char *term;
  float fval;
  int docid = 0;
  register char *pc;
  std::vector< std::pair< int, double > >* x; 
  std::string strbuf;
  char* buf;
  std::ifstream ifs;

  printf( "Reading test file \"%s\" ... \n", fname);

  ifs.open( fname );
  if ( ifs.fail() ) { printf( "Not found!\n" ); exit(1); }

  //  fp = fopen( fname, "r" );
  //  if ( fp==NULL ) { printf( "Not found!\n" ); exit(1); }

  /* process each line */
  while ( std::getline(ifs,strbuf) ) {
    line++;
    x = new std::vector< std::pair< int, double > >; 

    buf = new char[strbuf.size()+1];
    strcpy( buf, strbuf.c_str() );

    pc = buf;
    /* id 0 is  bias */
    x->push_back( std::pair<int,double>(0, 1.0) );

    /* read the class label */
    term = pc;
    pc = strchr( pc, ' ' );
    icat = atoi( term );

    Y[docid] = icat;

    /* read feature list */
    while ( *pc != '\0' ) {
        /* read feature id */
        pc++;
        if (*pc == '\n') break;
        iatt=0;
        while ( *pc != ':' ) { iatt = iatt*10+(*pc-48); pc++; }
        if ( iatt<0 ) fprintf( stderr,  "**** Error: nagatie value of xi=%d\n", iatt );
        if ( iatt>max_findex ) { max_findex = iatt; }
        /* read feature value */
        pc++;
        term = pc;
        pc = strchr( pc, ' ' );
        if ( pc == NULL ) {   
            fval = atof( term );
            x->push_back( std::pair<int,double>(iatt,fval) );
            break;
        } else {
            fval = atof( term );
            x->push_back( std::pair<int,double>(iatt,fval) );
	}
        if ( *(pc+1) == '\n') { break; }
    } /* end of while */

    docid++;
    if ( docid % 10000 == 0 ) {  fprintf( stderr, "*** %d data read\n", docid); }

    /* store a sample to X */
    X.push_back( *x );

  } /* end of while */

  ifs.close();
  //  fclose( fp );
  printf( "*** Largest feature ID = %d\n", max_findex );
  printf( "Done.\n" );
  return X.size();
}

double dot_product( int docid ) {
  double r=0;

  std::vector< std::pair< int, double > >::iterator iter;
  for( iter = X[docid].begin() ; iter != X[docid].end() ; iter++ ) {
    r += W[(*iter).first]*(*iter).second;
  }
  return r; 
}

void usage(char* command ) {
  fprintf( stderr, " (Usage) %s -t <test file> -m <model_file> [-r <result file>]\n", command );
}

void read_option (int argc, char *argv[]) {
  char option_char;

  /* options */
  /*   -m <model file> */
  /*   -t <test data file> */
  /*   -r <result file> */

  while ((option_char = getopt(argc, argv, "m:t:r:")) != EOF) {
    switch (option_char) {  
      case 'm': model_file = optarg; break;
      case 't': test_file = optarg; break;
      case 'r': result_file = optarg; break;
      case ':': usage( argv[0] ); exit(0);
      case '?': usage( argv[0] ); exit(0);
    }
  }
}

/**********************************************************************************/
